; Copyright (c) 2005, James Bailie <jimmy@mammothcheese.ca>.
; All rights reserved.
; 
; Redistribution and use in source and binary forms, with or without
; modification, are permitted provided that the following conditions are met:
; 
;     * Redistributions of source code must retain the above copyright
; notice, this list of conditions and the following disclaimer.
;     * Redistributions in binary form must reproduce the above copyright
; notice, this list of conditions and the following disclaimer in the
; documentation and/or other materials provided with the distribution.
;     * The name of James Bailie may not be used to endorse or promote 
; products derived from this software without specific prior written permission.
; 
; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS"
; AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
; IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
; ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
; LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
; CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
; SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
; INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
; CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
; ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
; POSSIBILITY OF SUCH DAMAGE.

; -----------------------------------------------------------------------

; THIS MODULE REQUIRES stacks.mm

; stacks.mm is not included with "#include" because there is a very good
; chance the program including this module also includes stacks.mm and the
; mm compiler is not smart enough to detect redundant inclusions.

; -------------------------------------------------------------------------

; cgi.mm -- helper functions for CGI programs.

; This module reserves the portion of the global namespace staring with
; cgi_ for its own use.

; The following six functions are defined:

; A CGI script must first invoke the following function to process
; the data submitted to it:  

; (cgi_get_data)
;
; "cgi_get_data" initializes a table named "cgi_data" with the decoded
; key/value pairs passed as arguments to the script, regardless of whether
; they were submitted by the "post" or "get" methods, if the data were
; encoded with x-www-form-urlencoding, which is the default encoding for
; forms.

; (cgi_send_headers title stylesheet script cookie favicon extras)
;
; send_headers sends the client the HTTP/1.1 and XHTML 1.0 Strict headers,
; and the sends the opening tag of the <body> element.  The title argument
; should be the text the user wishes to include in the <title></title>
; element.  The stylesheet argument should be the filename of any external
; stylesheet referenced by the page about to be created, and should contain
; only characters permissable in an XML attribute value (no quote
; characters or ampersands).  The script argument should be the filename of
; an external javascript program to specified as the value of the src 
; attribute of a <script /> tag.  The cookie argument is the desired body
; text of the Set-Cookie header, without the "Set-Cookie:" keyword and
; colon.  The favicon argument should be the filename of a favicon to be
; referenced in a <link...> element.  If the user does not wish to specify a
; title, stylesheet, script, a Set-Cookie header, or favicon, then the
; empty string may be passed for any of the arguments.  The last argument,
; extras, is either a boolean false value or a stack, containing strings, 
; each representing a complete HTTP/1.1 header to be added to the response
; to the client, including keyword and colon (eg. "Expires: Tue, 10 Aug
; 2004 10:22:31 GMT").

; (cgi_get_cookie)
;
; "cgi_get_cookie" initializes a new table bound to cgi_cookie comprising
; the key/value pairs included in the Cookie header passed to the script.
; It returns 1 if a cookie was found and a table created, or 0 if no Cookie
; header was present in the request.

; (cgi_make_cookie pairs path ttl)
;
; make_cookie returns a string representing the complete text of a
; Set-Cookie header.  The string may be passed as the "cookie" argument to
; send_headers to set the cookie.  The pairs argument must be a table of
; the key/value pairs the user wishes to set in the cookie.  The path
; argument should be the path the cookie applies to.  The ttl argument
; should be the age at which the cookie expires, expressed in seconds.

; (cgi_encode string)
; (cgi_decode string)
;
; These pair of functions encode-to and decode-from x-www-form-urlencoding,
; They both accept a string as argument, and return a new string.
;
;---------------------------------------------------------------------------

(let ((reserved (concat "]$;:@&+!?=#{}/[^`~\"<>|\\" (char 9) (char 10) (char 13)))

      (r2 (regcomp " "))
      (r3 (regcomp "%"))
      (r4 (regcomp "\\\\"))

      (adjust (lambda (d)
                 (cond ((eq d "10") "A")
                       ((eq d "11") "B")
                       ((eq d "12") "C")
                       ((eq d "13") "D")
                       ((eq d "14") "E")
                       ((eq d "15") "F")
                       (1 d))))

      (regexps (table)))

   (let ((rx (regcomp (concat "[" (substitute r4 "\\\\\\\\" reserved 1) "]"))))

      (apply_stack
         (lambda (x) 
            (let ((item (stack)))
               (push item (regcomp (if (or (eq x ">") (eq x "<")) x (concat "\\" x))))
               (push item (concat "%" (adjust (stringify (/ (code x) 16)))
                                      (adjust (stringify (% (code x) 16)))))
               (associate regexps x item)))

         (explode reserved))
      
      (setq cgi_encode_helper
         (lambda (m x)
            (if (not m)
               x

               (letn ((r (lookup regexps (index m 0)))
                      (s (substitute (index r 0) (index r 1) x 0)))

                  (cgi_encode_helper (matches rx s) s)))))

      (setq cgi_encode
         (lambda (x)
            (let ((s (substitute r3 "%25" x 0)))
               (substitute r2 "+" (cgi_encode_helper (matches rx s) s) 0))))))

(let ((regexps (table))
      (r0 (regcomp "%([0-9A-Fa-f][0-9A-Fa-f])"))
      (r1 (regcomp "\\+"))
      (r2 (regcomp ".")))

   (setq cgi_decode 
      (lambda (x)
         (cgi_decode_helper (substitute r1 " " x 0) (match r0 x) (matches r0 x))))

   (setq cgi_decode_helper
      (lambda (x indices encoding)
         (if (not encoding)
            x

            (let ((items (lookup regexps (index encoding 0))))

               (if items
                  (concat (substitute (index items 0) 
                                      (index items 1)
                                      (substring x 0 (index indices 1))
                                      1)
                          (if (eq (length x) (index indices 1))
                             ""
                             (let ((x (substring x (index indices 1) 0)))
                                (cgi_decode_helper x (match r0 x) (matches r0 x)))))
                  
                  (letn ((rx (regcomp (index encoding 0)))

                         (hex_digit_to_decimal
                            (lambda (x)
                               (let ((c (code (substitute r2 "\u\&" x))))
                                  (if (and (>= c 65) (<= c 70))
                                     (- c 55)
                                     (digitize x)))))

                         (digits (map_stack hex_digit_to_decimal (explode (index encoding 1))))
                         (converted (char (+ (* 16 (index digits 0)) (index digits 1))))
                         (new (stack)))

                     (push new rx)
                     (push new converted)
                     (associate regexps (index encoding 0) new)
                     (cgi_decode_helper x indices encoding))))))))

; Returns a complete Set-Cookie header.
; pairs is a table mapping keys to values.
; path is a string
; ttl can be a string or a number.

(setq cgi_make_cookie
   (lambda (pairs path ttl)
      (let ((cookie (cgi_make_cookie_helper pairs (keys pairs) "")))
         (concat "Set-Cookie: " cookie "Version=1; Path=" path "; Max-Age=" (stringify ttl)))))

(setq cgi_make_cookie_helper
   (lambda (pairs cookie_keys text)
      (if (not (used cookie_keys))
         text

         (let ((key (shift cookie_keys)))
            (cgi_make_cookie_helper pairs cookie_keys
               (concat text (stringify key) "=" (stringify (lookup pairs key)) "; "))))))
   
; Returns 1 if cookie was found, or 0 if no cookie is present in the HTTP
; headers.  If found, initializes a new table bound to cgi_cookie,
; containing the key/value pairs submitted.

(let ((leading_rx (regcomp "^[\b\t]+(.*)[\b\t]+$")))

   (setq cgi_get_cookie
      (lambda ()

         (let ((cookie (getenv "HTTP_COOKIE")))

            (if (not cookie)
               0
               
               (setq cgi_cookie (table))
      
               (apply_stack
                  (lambda (x) 
                     (associate cgi_cookie (substitute leading_rx "\1" (index x 0)) 
                                           (substitute leading_rx "\1" (index x 1))))
      
                  (map_stack
                     (lambda (x) (split "=" x))
                     (split ",;" cookie)))
      
               1)))))
   
(let ((ie_rx (regcomp "MSIE 6\.0")))

   (setq cgi_send_headers
      (lambda (title stylesheet script cookie favicon extras)
   
         ; HTTP headers.
   
         (print "Content-Type: text/html")
         (newline)
   
         ; Set-Cookie header.
   
         (when cookie
            (print cookie)
            (newline))
   
         ; Extra HTTP headers.
   
         (when extras
            (apply_stack (lambda (x) (print x) (newline)) extras))
   
         ; Header delimiting newline.
   
         (newline)
   
         ; XML headers.
   
         (let ((ua (getenv "HTTP_USER_AGENT")))
            (unless (and ua (match ie_rx ua))
               (print "<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>")
               (newline)
               (newline)))
         
         ; DOCTYPE document definition.
   
         (print "<!DOCTYPE html")
         (newline)
         (print "   PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\"")
         (newline)
         (print "   \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">")
         (newline)
         (newline)
   
         ; HTML header.
   
         (print "<html xmlns=\"http://www.w3.org/1999/xhtml\">")
         (newline)
         (newline)
   
         ; Head element.
   
         (print "<head>")
         (newline)
   
         (print "<title>" title "</title>")
         (newline)
   
         (when stylesheet
            (print "<link rel=\"stylesheet\" type=\"text/css\" media=\"screen\" href=\"" stylesheet "\" />")
            (newline))
   
         (when script
            (print "<script type=\"text/javascript\" src=\"" script "\"></script>")
            (newline))
   
         (when favicon
            (print "<link rel=\"shortcut icon\" href=\"" favicon "\" />")
            (newline))
   
         (print "</head>")
         (newline)
         (newline)
   
         ; Body element.
   
         (print "<body>")
         (newline))))

(setq cgi_get_data
   (lambda ()

      ; clear out any old parameters from the last invocation.

      (apply_stack (lambda (x) (dissociate cgi_data x)) (keys cgi_data))

      ; Get the new parameters.

      (when (eq (getenv "REQUEST_METHOD") "GET")

         (let ((d (if (eq (getenv "REQUEST_METHOD") "GET") 
                      (getenv "QUERY_STRING")
                      (readchars (digitize (or (getenv "CONTENT_LENGTH") ""))))))
   
            (if (not d)
               0
               
               (apply_stack
                  (lambda (x) 
                     (associate cgi_data 
                        (cgi_decode (index x 0))
                        (if (eq (used x) 2) (cgi_decode (index x 1)) "")))

                  (map_stack 
                     (lambda (x) (split "=" x))
                     (split "&" d)))

               1)))))

(setq cgi_data (table))
